## $Id: interfaces.py 13561 2015-12-22 07:25:36Z henrik $
##
## Copyright (C) 2011 Uli Fouquet & Henrik Bettermann
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
##
"""Customized interfaces of the university application package.
"""

from zope import schema
from zope.component import getUtility
from zope.schema import getFields
from waeup.kofa.applicants.interfaces import (
    contextual_reg_num_source,
    IApplicantBaseData,
    AppCatCertificateSource, CertificateSource)
from waeup.kofa.schoolgrades import ResultEntryField
from waeup.kofa.interfaces import (
    SimpleKofaVocabulary,
    academic_sessions_vocab,
    validate_email,
    SubjectSource,
    IKofaUtils)
from waeup.kofa.schema import FormattedDate, TextLineChoice
from waeup.kofa.students.vocabularies import nats_vocab, GenderSource
from kofacustom.nigeria.interfaces import (
    LGASource, high_qual, high_grade, exam_types)
from kofacustom.nigeria.interfaces import MessageFactory as _
from kofacustom.nigeria.payments.interfaces import INigeriaOnlinePayment

programme_types_vocab = SimpleKofaVocabulary(
    (_('Post UTME'), 'putme'),
    (_('Post DE'), 'pude'),
    (_('not applicable'), 'na'),
    )

jambsubjects = SimpleKofaVocabulary(
    (_('Use of English'),'english_language'),
    (_('Agricultural Science'),'agricultural_science'),
    (_('Arabic'),'arabic'),
    (_('Biology'),'biology'),
    (_('Book Keeping'),'book_keeping'),
    (_('Chemistry'),'chemistry'),
    (_('Christian Religious Studies'),'christian_religious_studies'),
    (_('Commerce'),'commerce'),
    (_('Economics'),'economics'),
    (_('Financial Accounting'),'financial_accounting'),
    (_('Fine Art'),'fine_art'),
    (_('Food and Nutrition'),'food_and_nutrition'),
    (_('French'),'french'),
    (_('Geography'),'geography'),
    (_('German'),'german'),
    (_('Government'),'government'),
    (_('Hausa'),'hausa'),
    (_('Home Economics'),'home_economics'),
    (_('History'),'history'),
    (_('Igbo'),'igbo'),
    (_('Literature in English'),'literature_in_english'),
    (_('Literature in Nigerian Languages'),'literature_in_nigerian_languages'),
    (_('Mathematics'),'mathematics'),
    (_('Music'),'music'),
    (_('Physics'),'physics'),
    (_('Yoruba'),'yoruba'),
    )

# Define a validation method for jamb subjects
class NotExactNumberOfItems(schema.ValidationError):
    __doc__ = u"Not exactly 4 items selected."

def four_items(value):
    if len(value) and len(value) != 4:
        raise NotExactNumberOfItems(value)
    return True

class JAMBSubjectSource(SubjectSource):
    """A source for school subjects used in exam documentation.
    """

    def getTitle(self, value):
        subjects_dict = getUtility(IKofaUtils).EXAM_SUBJECTS_DICT
        return "%s" % subjects_dict[value]

# Fields to be omitted in all display forms. course_admitted is
# rendered separately.

OMIT_DISPLAY_FIELDS = ('locked', 'course_admitted',
    'result_uploaded', 'suspended', 'special_application')

# UG students are all undergraduate students.
UG_OMIT_DISPLAY_FIELDS = OMIT_DISPLAY_FIELDS + (
    'jamb_subjects_list', 'programme_type')
UG_OMIT_PDF_FIELDS = UG_OMIT_DISPLAY_FIELDS + ('phone',)
UG_OMIT_MANAGE_FIELDS = (
    'special_application',
    'jamb_subjects_list',
    'programme_type')
UG_OMIT_EDIT_FIELDS = UG_OMIT_MANAGE_FIELDS + OMIT_DISPLAY_FIELDS + (
    'student_id',
    'notice',
    'screening_score',
    'screening_venue',
    'screening_date',
    'jamb_age',
    'jamb_subjects',
    'jamb_score',
    'jamb_reg_number',
    'aggregate')

# CBT is a subgroup of UG with the same interface.
CBT_OMIT_FIELDS = (
    'hq_type', 'hq_matric_no',
    'hq_degree', 'hq_school',
    'hq_session', 'hq_disc',
    'aggregate', 'jamb_subjects',
    'programme_type')
CBT_OMIT_DISPLAY_FIELDS = OMIT_DISPLAY_FIELDS + CBT_OMIT_FIELDS
CBT_OMIT_MANAGE_FIELDS = CBT_OMIT_FIELDS + ('special_application',)
CBT_OMIT_EDIT_FIELDS = OMIT_DISPLAY_FIELDS + CBT_OMIT_FIELDS + (
    'special_application',
    'student_id',
    'notice',
    'screening_score',
    'screening_venue',
    'screening_date',
    #'jamb_age',
    #'jamb_subjects',
    #'jamb_score',
    #'jamb_reg_number',
    )
CBT_OMIT_PDF_FIELDS = CBT_OMIT_DISPLAY_FIELDS + ('phone',)

# AFFIL is a subgroup of UG with the same interface.
AFFIL_OMIT_FIELDS = (
    'hq_type', 'hq_matric_no',
    'hq_degree', 'hq_school',
    'hq_session', 'hq_disc',
    'aggregate','jamb_subjects',)
AFFIL_OMIT_DISPLAY_FIELDS = OMIT_DISPLAY_FIELDS + AFFIL_OMIT_FIELDS
AFFIL_OMIT_MANAGE_FIELDS = AFFIL_OMIT_FIELDS + ('special_application',)
AFFIL_OMIT_EDIT_FIELDS = OMIT_DISPLAY_FIELDS + AFFIL_OMIT_FIELDS + (
    'special_application',
    'student_id',
    'notice',
    'screening_score',
    'screening_venue',
    'screening_date',
    )
AFFIL_OMIT_PDF_FIELDS = AFFIL_OMIT_DISPLAY_FIELDS + ('phone',)

# PUTME is a subgroup of UG with the same interface.
PUTME_OMIT_FIELDS = (
    'hq_type', 'hq_matric_no',
    'hq_degree', 'hq_school',
    'hq_session', 'hq_disc',
    'jamb_subjects_list', 'programme_type')
PUTME_OMIT_DISPLAY_FIELDS = OMIT_DISPLAY_FIELDS + PUTME_OMIT_FIELDS
PUTME_OMIT_MANAGE_FIELDS = UG_OMIT_MANAGE_FIELDS + PUTME_OMIT_FIELDS
PUTME_OMIT_EDIT_FIELDS = UG_OMIT_EDIT_FIELDS + PUTME_OMIT_FIELDS + (
    'firstname', 'middlename', 'lastname', 'sex',
    'course1', 'lga')
PUTME_OMIT_PDF_FIELDS = PUTME_OMIT_DISPLAY_FIELDS + ('phone',)
PUTME_OMIT_RESULT_SLIP_FIELDS = PUTME_OMIT_DISPLAY_FIELDS + (
    'phone',
    'date_of_birth', 'sex',
    'nationality', 'lga', #'perm_address',
    'course2', 'screening_venue',
    'screening_date')

# PUDE is a subgroup of UG with the same interface.
PUDE_OMIT_FIELDS = (
    'jamb_subjects',
    'jamb_score',
    'jamb_age',
    'aggregate',
    'jamb_subjects_list',
    'programme_type')
PUDE_OMIT_DISPLAY_FIELDS = OMIT_DISPLAY_FIELDS + PUDE_OMIT_FIELDS
PUDE_OMIT_MANAGE_FIELDS = UG_OMIT_MANAGE_FIELDS + PUDE_OMIT_FIELDS
PUDE_OMIT_EDIT_FIELDS = set(UG_OMIT_EDIT_FIELDS + PUDE_OMIT_FIELDS + (
    'firstname', 'middlename', 'lastname', 'sex',
    'course1', 'lga'))
PUDE_OMIT_PDF_FIELDS = PUDE_OMIT_DISPLAY_FIELDS + ('phone',)
PUDE_OMIT_RESULT_SLIP_FIELDS = PUDE_OMIT_DISPLAY_FIELDS + (
    'phone',
    'date_of_birth', 'sex',
    'nationality', 'lga', #'perm_address',
    'course2', 'screening_venue',
    'screening_date')

# PG has its own interface
PG_OMIT_DISPLAY_FIELDS = OMIT_DISPLAY_FIELDS
PG_OMIT_PDF_FIELDS = PG_OMIT_DISPLAY_FIELDS + ('phone',)
PG_OMIT_MANAGE_FIELDS = ('special_application',)
PG_OMIT_EDIT_FIELDS = PG_OMIT_MANAGE_FIELDS + PG_OMIT_DISPLAY_FIELDS + (
    'student_id', 'notice',
    'screening_score', 'screening_venue',
    'screening_date',)

class INigeriaUGApplicant(IApplicantBaseData):
    """An undergraduate applicant.

    This interface defines the least common multiple of all fields
    in ug application forms. In customized forms, fields can be excluded by
    adding them to the UG_OMIT* tuples.
    """

    nationality = schema.Choice(
        source = nats_vocab,
        title = _(u'Nationality'),
        required = False,
        )
    lga = schema.Choice(
        source = LGASource(),
        title = _(u'State/LGA (Nigerians only)'),
        required = False,
        )
    #perm_address = schema.Text(
    #    title = _(u'Permanent Address'),
    #    required = False,
    #    )
    course1 = schema.Choice(
        title = _(u'1st Choice Course of Study'),
        source = AppCatCertificateSource(),
        required = True,
        )
    course2 = schema.Choice(
        title = _(u'2nd Choice Course of Study'),
        source = AppCatCertificateSource(),
        required = False,
        )

    programme_type = schema.Choice(
        title = _(u'Programme Type'),
        vocabulary = programme_types_vocab,
        required = False,
        )

    hq_type = schema.Choice(
        title = _(u'Qualification Obtained'),
        required = False,
        readonly = False,
        vocabulary = high_qual,
        )
    hq_matric_no = schema.TextLine(
        title = _(u'Former Matric Number'),
        required = False,
        readonly = False,
        )
    hq_degree = schema.Choice(
        title = _(u'Class of Degree'),
        required = False,
        readonly = False,
        vocabulary = high_grade,
        )
    hq_school = schema.TextLine(
        title = _(u'Institution Attended'),
        required = False,
        readonly = False,
        )
    hq_session = schema.TextLine(
        title = _(u'Years Attended'),
        required = False,
        readonly = False,
        )
    hq_disc = schema.TextLine(
        title = _(u'Discipline'),
        required = False,
        readonly = False,
        )
    jamb_subjects = schema.Text(
        title = _(u'Subjects and Scores'),
        required = False,
        )
    jamb_subjects_list = schema.List(
        title = _(u'JAMB Subjects'),
        required = False,
        default = [],
        value_type = schema.Choice(
            vocabulary = jambsubjects
            #source = JAMBSubjectSource(),
            ),
        )
    jamb_score = schema.Int(
        title = _(u'Total JAMB Score'),
        required = False,
        )
    #jamb_age = schema.Int(
    #    title = _(u'Age (provided by JAMB)'),
    #    required = False,
    #    )
    jamb_reg_number = schema.TextLine(
        title = _(u'JAMB Registration Number'),
        required = False,
        )
    notice = schema.Text(
        title = _(u'Notice'),
        required = False,
        )
    screening_venue = schema.TextLine(
        title = _(u'Screening Venue'),
        required = False,
        )
    screening_date = schema.TextLine(
        title = _(u'Screening Date'),
        required = False,
        )
    screening_score = schema.Int(
        title = _(u'Screening Score (%)'),
        required = False,
        )
    aggregate = schema.Int(
        title = _(u'Aggregate Score (%)'),
        description = _(u'(average of relative JAMB and PUTME scores)'),
        required = False,
        )
    result_uploaded = schema.Bool(
        title = _(u'Result uploaded'),
        default = False,
        required = False,
        )
    student_id = schema.TextLine(
        title = _(u'Student Id'),
        required = False,
        readonly = False,
        )
    course_admitted = schema.Choice(
        title = _(u'Admitted Course of Study'),
        source = CertificateSource(),
        required = False,
        )
    locked = schema.Bool(
        title = _(u'Form locked'),
        default = False,
        required = False,
        )

INigeriaUGApplicant[
    'locked'].order =  IApplicantBaseData['suspended'].order
INigeriaUGApplicant[
    'result_uploaded'].order =  INigeriaUGApplicant['suspended'].order

class INigeriaPGApplicant(IApplicantBaseData):
    """A postgraduate applicant.

    This interface defines the least common multiple of all fields
    in pg application forms. In customized forms, fields can be excluded by
    adding them to the PG_OMIT* tuples.
    """

    nationality = schema.Choice(
        source = nats_vocab,
        title = _(u'Nationality'),
        required = True,
        )
    lga = schema.Choice(
        source = LGASource(),
        title = _(u'State/LGA (Nigerians only)'),
        required = False,
        )
    #perm_address = schema.Text(
    #    title = _(u'Permanent Address'),
    #    required = False,
    #    )
    course1 = schema.Choice(
        title = _(u'1st Choice Course of Study'),
        source = AppCatCertificateSource(),
        required = True,
        )
    course2 = schema.Choice(
        title = _(u'2nd Choice Course of Study'),
        source = AppCatCertificateSource(),
        required = False,
        )
    hq_type = schema.Choice(
        title = _(u'Qualification Obtained'),
        required = False,
        readonly = False,
        vocabulary = high_qual,
        )
    hq_fname = schema.TextLine(
        title = _(u'Full Name'),
        required = False,
        readonly = False,
        )
    hq_matric_no = schema.TextLine(
        title = _(u'Former Matric Number'),
        required = False,
        readonly = False,
        )
    hq_degree = schema.Choice(
        title = _(u'Class of Degree'),
        required = False,
        readonly = False,
        vocabulary = high_grade,
        )
    hq_school = schema.TextLine(
        title = _(u'Institution Attended'),
        required = False,
        readonly = False,
        )
    hq_session = schema.TextLine(
        title = _(u'Years Attended'),
        required = False,
        readonly = False,
        )
    hq_disc = schema.TextLine(
        title = _(u'Discipline'),
        required = False,
        readonly = False,
        )
    fst_sit_fname = schema.TextLine(
        title = _(u'Full Name'),
        required = False,
        readonly = False,
        )
    fst_sit_no = schema.TextLine(
        title = _(u'Exam Number'),
        required = False,
        readonly = False,
        )
    fst_sit_date = FormattedDate(
        title = _(u'Exam Date'),
        required = False,
        readonly = False,
        show_year = True,
        )
    fst_sit_type = schema.Choice(
        title = _(u'Exam Type'),
        required = False,
        readonly = False,
        vocabulary = exam_types,
        )
    fst_sit_results = schema.List(
        title = _(u'Exam Results'),
        value_type = ResultEntryField(),
        required = False,
        readonly = False,
        default = [],
        )
    scd_sit_fname = schema.TextLine(
        title = _(u'Full Name'),
        required = False,
        readonly = False,
        )
    scd_sit_no = schema.TextLine(
        title = _(u'Exam Number'),
        required = False,
        readonly = False,
        )
    scd_sit_date = FormattedDate(
        title = _(u'Exam Date'),
        required = False,
        readonly = False,
        show_year = True,
        )
    scd_sit_type = schema.Choice(
        title = _(u'Exam Type'),
        required = False,
        readonly = False,
        vocabulary = exam_types,
        )
    scd_sit_results = schema.List(
        title = _(u'Exam Results'),
        value_type = ResultEntryField(),
        required = False,
        readonly = False,
        default = [],
        )
    # Replaced by first and second sitting
    #pp_school = schema.Choice(
    #    title = _(u'Qualification Obtained'),
    #    required = False,
    #    readonly = False,
    #    vocabulary = exam_types,
    #    )
    presently_inst = schema.TextLine(
        title = _(u'If yes, name of institution'),
        required = False,
        readonly = False,
        )
    nysc_year = schema.Int(
        title = _(u'Nysc Year'),
        required = False,
        readonly = False,
        )
    nysc_lga = schema.Choice(
        source = LGASource(),
        title = _(u'Nysc Location'),
        required = False,
        )
    employer = schema.TextLine(
        title = _(u'Employer'),
        required = False,
        readonly = False,
        )
    emp_position = schema.TextLine(
        title = _(u'Employer Position'),
        required = False,
        readonly = False,
        )
    emp_start = FormattedDate(
        title = _(u'Start Date'),
        required = False,
        readonly = False,
        show_year = True,
        )
    emp_end = FormattedDate(
        title = _(u'End Date'),
        required = False,
        readonly = False,
        show_year = True,
        )
    emp_reason = schema.TextLine(
        title = _(u'Reason for Leaving'),
        required = False,
        readonly = False,
        )
    employer2 = schema.TextLine(
        title = _(u'2nd Employer'),
        required = False,
        readonly = False,
        )
    emp2_position = schema.TextLine(
        title = _(u'2nd Employer Position'),
        required = False,
        readonly = False,
        )
    emp2_start = FormattedDate(
        title = _(u'Start Date'),
        required = False,
        readonly = False,
        show_year = True,
        )
    emp2_end = FormattedDate(
        title = _(u'End Date'),
        required = False,
        readonly = False,
        show_year = True,
        )
    emp2_reason = schema.TextLine(
        title = _(u'Reason for Leaving'),
        required = False,
        readonly = False,
        )
    notice = schema.Text(
        title = _(u'Notice'),
        required = False,
        readonly = False,
        )
    screening_venue = schema.TextLine(
        title = _(u'Screening Venue'),
        required = False,
        )
    screening_date = schema.TextLine(
        title = _(u'Screening Date'),
        required = False,
        )
    screening_score = schema.Int(
        title = _(u'Screening Score (%)'),
        required = False,
        )
    student_id = schema.TextLine(
        title = _(u'Student Id'),
        required = False,
        readonly = False,
        )
    course_admitted = schema.Choice(
        title = _(u'Admitted Course of Study'),
        source = CertificateSource(),
        required = False,
        readonly = False,
        )
    locked = schema.Bool(
        title = _(u'Form locked'),
        default = False,
        required = False,
        )

INigeriaPGApplicant[
    'locked'].order =  IApplicantBaseData['suspended'].order

# PRE is used by Uniben
#  med: "it is as named... PRE - DEGREE... much like a 1 year diploma programme"
PRE_OMIT_FIELDS = tuple([
    i for i in getFields(INigeriaPGApplicant).keys()
        if i[:3] in ('hq_', 'emp', 'nys')])
PRE_OMIT_DISPLAY_FIELDS = OMIT_DISPLAY_FIELDS + PRE_OMIT_FIELDS
PRE_OMIT_MANAGE_FIELDS = PG_OMIT_MANAGE_FIELDS + PRE_OMIT_FIELDS
PRE_OMIT_EDIT_FIELDS = set(PG_OMIT_EDIT_FIELDS + PRE_OMIT_FIELDS)
PRE_OMIT_PDF_FIELDS = PRE_OMIT_DISPLAY_FIELDS
PRE_OMIT_RESULT_SLIP_FIELDS = PRE_OMIT_DISPLAY_FIELDS

class INigeriaApplicant(INigeriaUGApplicant, INigeriaPGApplicant):
    """An interface for both types of applicants.

    Attention: The INigeriaPGApplicant field seetings will be overwritten
    by INigeriaPGApplicant field settings. If a field is defined
    in both interfaces zope.schema validates only against the
    constraints in INigeriaUGApplicant. This does not affect the forms
    since they are build on either INigeriaUGApplicant or INigeriaPGApplicant.
    """

    def writeLogMessage(view, comment):
        """Adds an INFO message to the log file
        """

    def createStudent():
        """Create a student object from applicant data
        and copy applicant object.
        """

class INigeriaUGApplicantEdit(INigeriaUGApplicant):
    """An undergraduate applicant interface for edit forms.

    Here we can repeat the fields from base data and set the
    `required` and `readonly` attributes to True to further restrict
    the data access. Or we can allow only certain certificates to be
    selected by choosing the appropriate source.

    We cannot omit fields here. This has to be done in the
    respective form page.
    """

    email = schema.ASCIILine(
        title = _(u'Email Address'),
        required = True,
        constraint=validate_email,
        )
    date_of_birth = FormattedDate(
        title = _(u'Date of Birth'),
        required = True,
        show_year = True,
        )
    jamb_subjects_list = schema.List(
        title = _(u'JAMB Subjects'),
        description = _(u'Select four subjects.'),
        required = True,
        constraint = four_items,
        value_type = schema.Choice(
            vocabulary = jambsubjects
            #source = JAMBSubjectSource(),
            ),
        )

INigeriaUGApplicantEdit[
    'date_of_birth'].order = INigeriaUGApplicant['date_of_birth'].order
INigeriaUGApplicantEdit[
    'email'].order = INigeriaUGApplicant['email'].order
INigeriaUGApplicantEdit[
    'jamb_subjects_list'].order = INigeriaUGApplicant['jamb_subjects_list'].order

class INigeriaPGApplicantEdit(INigeriaPGApplicant):
    """A postgraduate applicant interface for editing.

    Here we can repeat the fields from base data and set the
    `required` and `readonly` attributes to True to further restrict
    the data access. Or we can allow only certain certificates to be
    selected by choosing the appropriate source.

    We cannot omit fields here. This has to be done in the
    respective form page.
    """

    email = schema.ASCIILine(
        title = _(u'Email Address'),
        required = True,
        constraint=validate_email,
        )
    date_of_birth = FormattedDate(
        title = _(u'Date of Birth'),
        required = True,
        show_year = True,
        )

INigeriaPGApplicantEdit[
    'date_of_birth'].order =  INigeriaPGApplicant['date_of_birth'].order
INigeriaPGApplicantEdit[
    'email'].order =  INigeriaPGApplicant['email'].order

class INigeriaApplicantOnlinePayment(INigeriaOnlinePayment):
    """An applicant payment via payment gateways.
    """

class IPUTMEApplicantEdit(INigeriaUGApplicant):
    """An undergraduate applicant interface for editing.

    Here we can repeat the fields from base data and set the
    `required` and `readonly` attributes to True to further restrict
    the data access. Or we can allow only certain certificates to be
    selected by choosing the appropriate source.

    We cannot omit fields here. This has to be done in the
    respective form page.
    """
    email = schema.ASCIILine(
        title = _(u'Email Address'),
        required = True,
        constraint=validate_email,
        )
    date_of_birth = FormattedDate(
        title = _(u'Date of Birth'),
        required = True,
        show_year = True,
        )
    nationality = schema.Choice(
        source = nats_vocab,
        title = _(u'Nationality'),
        required = True,
        )

IPUTMEApplicantEdit[
    'date_of_birth'].order =  INigeriaUGApplicant['date_of_birth'].order
IPUTMEApplicantEdit[
    'email'].order =  INigeriaUGApplicant['email'].order
IPUTMEApplicantEdit[
    'nationality'].order =  INigeriaUGApplicant['nationality'].order

class INigeriaApplicantUpdateByRegNo(INigeriaApplicant):
    """Representation of an applicant.

    Skip regular reg_number validation if reg_number is used for finding
    the applicant object.
    """
    reg_number = schema.TextLine(
        title = u'Registration Number',
        required = False,
        )
